import uuid


class Invoice:
    _details: dict = {}
    transaction_id: str
    amount: int
    driver: str

    def __init__(self):
        self.create_uuid()

    def create_uuid(self, uuid_param=None):
        if not uuid_param:
            uuid_param = str(uuid.uuid4())

        self.uuid = uuid_param

    def get_uuid(self):
        return self.uuid

    def via(self, driver):
        self.driver = driver

    def set_amount(self, amount):
        if not isinstance(amount, int):
            raise Exception('Amount value should be a number (integer).')
        self.amount = amount

    def set_transaction_id(self, transaction_id):
        self.transaction_id = transaction_id

    def get_transaction_id(self):
        return self.transaction_id

    def set_details(self, key, value):
        self._details[key] = value

    def get_details(self) -> dict:
        return self._details

    def get_detail(self, key) -> str:
        return self._details.get(key)



from abc import ABC, abstractmethod

from django_payment_gateways.modules.invoice import Invoice


class Driver(ABC):

    @abstractmethod
    def __init__(self, invoice: Invoice, settings: dict):
        pass

    def set_amount(self, amount):
        self.amount = amount

    def set_invoice(self, invoice: Invoice):
        self.invoice = invoice

    @abstractmethod
    def purchase(self):
        pass

    @abstractmethod
    def pay(self):
        pass

    @abstractmethod
    def verify(self, request):
        pass



from django.utils import timezone


class Receipt:
    _reference_id: str = None
    _driver: str = None
    _date = None

    def __init__(self, driver, reference_id):
        self._driver = driver
        self._reference_id = reference_id
        self._date = timezone.now()

    def get_driver(self) -> str:
        return self._driver

    def get_reference_id(self) -> str:
        return self._reference_id

    def get_date(self):
        return self._date



from django_payment_gateways.modules.abstracts import Receipt
from django_payment_gateways.modules.abstracts.driver import Driver
from django_payment_gateways.modules.drivers.parsian.strategies.normal import Normal
from django_payment_gateways.modules.drivers.parsian.strategies.sandbox import Sandbox
from django_payment_gateways.modules.invoice import Invoice


class Parsian(Driver):
    strategies = {
        'normal': Normal,
        'sandbox': Sandbox
    }

    def __init__(self, invoice: Invoice, settings: dict):
        self.invoice = invoice
        self.settings = settings
        self.strategy = self._get_fresh_strategy_instance(invoice, settings)

    def purchase(self):
        return self.strategy.purchase()

    def pay(self):
        return self.strategy.pay()

    def verify(self, request):
        return self.strategy.verify(request)

    def _get_fresh_strategy_instance(self, invoice: Invoice, settings: dict):
        strategy = self.strategies[self._get_mode()]
        return strategy(invoice, settings)

    def _get_mode(self):
        return str(self.settings.get('mode')).lower()



from django.shortcuts import redirect
from zeep import Client

from django_payment_gateways.exceptions import PurchaseFailedException, InvalidPaymentException
from django_payment_gateways.modules.abstracts import Driver, Receipt
from django_payment_gateways.modules.invoice import Invoice


class Normal(Driver):
    _invoice: Invoice
    _settings: dict

    def __init__(self, invoice: Invoice, settings: dict):
        self._invoice = invoice
        self._settings = settings

    def purchase(self) -> str:
        data = {
            "LoginAccount": self._settings.get("merchantId"),
            "Amount": self._invoice.amount * 10,
            "AdditionalData": self._invoice.get_detail("description") or self._settings.get("description"),
            "OrderId": self._invoice.get_uuid(),
            "CallBackUrl": self._settings.get("callbackUrl")
        }

        result = self._get_client(self.get_purchase_url()).service.SalePaymentRequest(requestData=data)
        token = result.Token
        status = int(result.Status)

        if token and status == 0:
            self._invoice.set_transaction_id(token)
            print(token)
            return self._invoice.transaction_id
        else:
            message = self.translate_status(str(status))
            raise PurchaseFailedException(message, str(status))

    def pay(self):
        transaction_id = self._invoice.transaction_id
        payment_url = self.get_payment_url()
        pay_url = f'{payment_url}?Token={transaction_id}'
        return redirect(pay_url)

    def verify(self, request):
        status = int(request.POST.get('status'))
        token = self._invoice.transaction_id or request.POST.get('Token')

        if status != 0 or not token:
            raise InvalidPaymentException('عملیات پرداخت توسط کاربر لغو شد.', status)

        data = {
            'LoginAccount': self._settings.get('merchantId'),
            'Token': token
        }
        result = self._get_client(self.get_verification_url()).service.ConfirmPayment(requestData=data)

        status = int(result.Status)
        rrn = str(result.RRN)
        if status == -138:
            raise InvalidPaymentException(
                "این تراکنش پرداخت شده بوده است"
            )

        if status != 0:
            message = self.translate_status(str(status))
            raise InvalidPaymentException(message, status)

        if status == 0:
            # successful
            return self.create_receipt(rrn)
        else:
            message = self.translate_status(str(status))
            raise InvalidPaymentException(message, status)

    def get_purchase_url(self) -> str:
        return self._settings.get('apiPurchaseUrl')

    def get_payment_url(self) -> str:
        return self._settings.get('apiPaymentUrl')

    def get_verification_url(self) -> str:
        return self._settings.get('apiVerificationUrl')

    def _get_client(self, url: str):
        return Client(url)

    def translate_status(self, status: str) -> str:
        translations = {
            "-32768": "UnknownError",
            "-1552": "PaymentRequestIsNotEligibleToReversal",
            "-1551": "PaymentRequestIsAlreadyReversed",
            "-1550": "PaymentRequestStatusIsNotReversible",
            "-1549": "MaxAllowedTimeToReversalHasExceeded",
            "-1548": "BillPaymentRequestServiceFailed",
            "-1540": "InvalidConfirmRequestService",
            "-1536": "TopupChargeServiceTopupChargeRequestFailed",
            "-1533": "PaymentIsAlreadyConfirmed",
            "-1532": "MerchantHasConfirmedPaymentRequest",
            "-1531": "CannotConfirmNonSuccessfulPayment",
            "-1530": "MerchantConfirmPaymentRequestAccessViolated",
            "-1528": "ConfirmPaymentRequestInfoNotFound",
            "-1527": "CallSalePaymentRequestServiceFailed",
            "-1507": "ReversalCompleted",
            "-1505": "PaymentConfirmRequested",
            "-138": "CanceledByUser",
            "-132": "InvalidMinimumPaymentAmount",
            "-131": "InvalidToken",
            "-130": "TokenIsExpired",
            "-128": "InvalidIpAddressFormat",
            "-127": "InvalidMerchantIp",
            "-126": "InvalidMerchantPin",
            "-121": "InvalidStringIsNumeric",
            "-120": "InvalidLength",
            "-119": "InvalidOrganizationId",
            "-118": "ValueIsNotNumeric",
            "-117": "LengthIsLessOfMinimum",
            "-116": "LengthIsMoreOfMaximum",
            "-115": "InvalidPayId",
            "-114": "InvalidBillId",
            "-113": "ValueIsNull",
            "-112": "OrderIdDuplicated",
            "-111": "InvalidMerchantMaxTransAmount",
            "-108": "ReverseIsNotEnabled",
            "-107": "AdviceIsNotEnabled",
            "-106": "ChargeIsNotEnabled",
            "-105": "TopupIsNotEnabled",
            "-104": "BillIsNotEnabled",
            "-103": "SaleIsNotEnabled",
            "-102": "ReverseSuccessful",
            "-101": "MerchantAuthenticationFailed",
            "-100": "MerchantIsNotActive",
            "-1": "Server Error",
            "0": "Successful",
            "1": "Refer To Card Issuer Decline",
            "2": "Refer To Card Issuer Special Conditions",
            "3": "Invalid Merchant",
            "5": "Do Not Honour",
            "6": "Error",
            "8": "Honour With Identification",
            "9": "Request In-progress",
            "10": "Approved For Partial Amount",
            "12": "Invalid Transaction",
            "13": "Invalid Amount",
            "14": "Invalid Card Number",
            "15": "No Such Issuer",
            "17": "Customer Cancellation",
            "20": "Invalid Response",
            "21": "No Action Taken",
            "22": "Suspected Malfunction",
            "30": "Format Error",
            "31": "Bank Not Supported By Switch",
            "32": "Completed Partially",
            "33": "Expired Card Pick Up",
            "38": "Allowable PIN Tries Exceeded Pick Up",
            "39": "No Credit Account",
            "40": "Requested Function is not supported",
            "41": "Lost Card",
            "43": "Stolen Card",
            "45": "Bill Can not Be Payed",
            "51": "No Sufficient Funds",
            "54": "Expired Account",
            "55": "Incorrect PIN",
            "56": "No Card Record",
            "57": "Transaction Not Permitted To CardHolder",
            "58": "Transaction Not Permitted To Terminal",
            "59": "Suspected Fraud-Decline",
            "61": "Exceeds Withdrawal Amount Limit",
            "62": "Restricted Card-Decline",
            "63": "Security Violation",
            "65": "Exceeds Withdrawal Frequency Limit",
            "68": "Response Received Too Late",
            "69": "Allowable Number Of PIN Tries Exceeded",
            "75": "PIN Reties Exceeds-Slm",
            "78": "Deactivated Card-Slm",
            "79": "Invalid Amount-Slm",
            "80": "Transaction Denied-Slm",
            "81": "Cancelled Card-Slm",
            "83": "Host Refuse-Slm",
            "84": "Issuer Down-Slm",
            "91": "Issuer Or Switch Is Inoperative",
            "92": "Not Found for Routing",
            "93": "Cannot Be Completed",
        }
        unknown_error = 'خطای ناشناخته رخ داده است.'
        return translations[status] if status in translations else unknown_error

    def create_receipt(self, reference_id: str) -> Receipt:
        return Receipt('parsian', reference_id)



from django.db import IntegrityError


class InsufficientBalance(IntegrityError):
    """Raised when a wallet has insufficient balance to
    run an operation.
    We're subclassing from :mod:`django.db.IntegrityError`
    so that it is automatically rolled-back during django's
    transaction lifecycle.
    """


class DriverNotFoundException(Exception):
    """Driver not selected or default driver does not exist."""


class InvoiceNotFoundException(Exception):
    """invoice not found"""


class PurchaseFailedException(Exception):
    """purchase failed"""


class InvalidPaymentException(Exception):
    """invalid payment"""



from importlib import import_module

from django.conf import settings

from django_payment_gateways.exceptions import DriverNotFoundException, InvoiceNotFoundException
from django_payment_gateways.modules.abstracts import Driver
from django_payment_gateways.modules.invoice import Invoice

gateway_cache = {}


class PaymentWorker:
    _driver_instance: Driver = None
    _invoice: Invoice = None

    def __init__(self):
        self.config = self._get_default_config()
        self._invoice = Invoice()
        self.via(self.get_default_driver())

    def _get_default_config(self) -> dict:
        return getattr(settings, 'PAYMENT_SETTINGS', {})

    def get_default_driver(self) -> str:
        return self.config['drivers']['default']

    def get_driver_config(self, driver) -> dict:
        return self.config['drivers'][driver]

    def via(self, driver):
        self.driver = driver
        self.validate_driver()
        self._invoice.via(driver)
        self.settings = self.get_driver_config(driver)

    def validate_driver(self):
        if not self.driver:
            raise DriverNotFoundException('Driver not selected or default driver does not exist.')

        if not self.config['drivers'][self.driver] or not self.config['map'][self.driver]:
            raise DriverNotFoundException('Driver not found in config file. Try updating the package.');

    def validate_invoice(self):
        if not self._invoice:
            raise InvoiceNotFoundException('Invoice not selected or does not exist.')

    def set_amount(self, amount):
        self._invoice.set_amount(amount)

    def set_invoice(self, invoice):
        self._invoice = invoice

    def set_callback(self, callback_url):
        self.settings['callbackUrl'] = callback_url

    def purchase(self, invoice: Invoice = None):
        if invoice:
            self._invoice = invoice

        self._driver_instance = self.get_fresh_driver_instance()

        transaction_id = self._driver_instance.purchase()
        return self._driver_instance, transaction_id

    def pay(self):
        self._driver_instance = self.get_driver_instance()
        self.validate_invoice()
        return self._driver_instance.pay()

    def verify(self, request):
        self._driver_instance = self.get_driver_instance()
        self.validate_invoice()
        receipt = self._driver_instance.verify(request)
        return receipt, self._driver_instance

    def get_driver_instance(self):
        if self._driver_instance:
            return self._driver_instance

        return self.get_fresh_driver_instance()

    def get_fresh_driver_instance(self):
        self.validate_driver()
        klass = self.import_driver()
        return klass(self._invoice, self.settings)

    def import_driver(self):
        """
        Return a gateway instance specified by `gateway` name.
        This caches gateway classes in a module-level dictionnary to avoid hitting
        the filesystem every time we require a gateway.
        Should the list of available gateways change at runtime, one should then
        invalidate the cache, the simplest of ways would be to:
        >>> gateway_cache = {}
        """
        # Is the class in the cache?
        clazz = gateway_cache.get(self.driver, None)
        if not clazz:
            # Let's actually load it (it's not in the cache)
            str_mod: str = self.config['map'][self.driver]
            module_name, class_name = str_mod.split(':')
            clazz = getattr(import_module(module_name), class_name)
            gateway_cache[self.driver] = clazz
        return clazz



from django.shortcuts import redirect
from zeep import Client

from django_payment_gateways.exceptions import PurchaseFailedException, InvalidPaymentException
from django_payment_gateways.modules.abstracts.driver import Driver
from django_payment_gateways.modules.abstracts.receipt import Receipt
from django_payment_gateways.modules.invoice import Invoice


class Sandbox(Driver):
    _invoice: Invoice
    _settings: dict

    def __init__(self, invoice: Invoice, settings: dict):
        self._invoice = invoice
        self._settings = settings

    def purchase(self) -> str:
        data = {
            'MerchantID': self._settings.get('merchantId'),
            'Amount': self._invoice.amount,
            'Description': self._invoice.get_detail('description') or self._settings.get('description'),
            'Mobile': self._invoice.get_detail('mobile') or self._settings.get('mobile'),
            'Email': self._invoice.get_detail('email') or self._settings.get('email'),
            'CallbackURL': self._settings.get('callbackUrl'),
        }
        result = self._get_client(self.get_purchase_url()).service.PaymentRequest(**data)

        if result.Status != 100 or not result.Authority:
            message = self.translate_status(result.Status)
            raise PurchaseFailedException(message, result.Status)

        self._invoice.set_transaction_id(result.Authority)
        return self._invoice.transaction_id

    def pay(self):
        transaction_id = self._invoice.transaction_id
        payment_url = self.get_payment_url()
        pay_url = payment_url + transaction_id
        return redirect(pay_url)

    def verify(self, request):
        status = request.GET.get('Status')
        if status != 'OK':
            raise InvalidPaymentException('عملیات پرداخت توسط کاربر لغو شد.', -22)

        authority = self._invoice.transaction_id or request.GET.get('Authority')

        data = {
            'MerchantID': self._settings.get('merchantId'),
            'Authority': authority,
            'Amount': self._invoice.amount,
        }
        result = self._get_client(self.get_verification_url()).service.PaymentVerification(**data)

        if result.Status != 100:
            message = self.translate_status(result.Status)
            raise InvalidPaymentException(message, result.Status)

        return self.create_receipt(result.RefID)

    def get_purchase_url(self) -> str:
        return self._settings.get('sandboxApiPurchaseUrl')

    def get_payment_url(self) -> str:
        return self._settings.get('sandboxApiPaymentUrl')

    def get_verification_url(self) -> str:
        return self._settings.get('sandboxApiVerificationUrl')

    def _get_client(self, url: str):
        return Client(url)

    def translate_status(self, status: str) -> str:
        translations = {
            "-1": "اطلاعات ارسال شده ناقص است.",
            "-2": "IP و يا مرچنت كد پذيرنده صحيح نيست",
            "-3": "با توجه به محدوديت هاي شاپرك امكان پرداخت با رقم درخواست شده ميسر نمي باشد",
            "-4": "سطح تاييد پذيرنده پايين تر از سطح نقره اي است.",
            "-11": "درخواست مورد نظر يافت نشد.",
            "-12": "امكان ويرايش درخواست ميسر نمي باشد.",
            "-21": "هيچ نوع عمليات مالي براي اين تراكنش يافت نشد",
            "-22": "تراكنش نا موفق ميباشد",
            "-33": "رقم تراكنش با رقم پرداخت شده مطابقت ندارد",
            "-34": "سقف تقسيم تراكنش از لحاظ تعداد يا رقم عبور نموده است",
            "-40": "اجازه دسترسي به متد مربوطه وجود ندارد.",
            "-41": "اطلاعات ارسال شده مربوط به AdditionalData غيرمعتبر ميباشد.",
            "-42": "مدت زمان معتبر طول عمر شناسه پرداخت بايد بين 30 دقيه تا 45 روز مي باشد.",
            "-54": "درخواست مورد نظر آرشيو شده است",
            "101": "عمليات پرداخت موفق بوده و قبلا PaymentVerification تراكنش انجام شده است.",
        }
        unknown_error = 'خطای ناشناخته رخ داده است.'
        return translations[status] if status in translations else unknown_error

    def create_receipt(self, reference_id: str) -> Receipt:
        return Receipt('zarinpal', reference_id)
